/******************************************************************************/
/* (c) 2006 jamie twycross, jpt AT cs.nott.ac.uk                              */
/* released under gnu gpl v2                                                  */
/* example two cell algorithm                                                 */
/******************************************************************************/

#include <tissue/tissue.h>
#include <tissue/syscall.h>

#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <asm/unistd.h>
#include <math.h>

/* model parameters */
unsigned int max_antigen = 1000;
unsigned int max_cytokines = 2;
unsigned int max_cells = 100;
unsigned int cell_update_rate = 100000;
unsigned int num_cells_1 = 50;
unsigned int num_antigen_1 = 100;
unsigned int num_antigen_receptors_1 = 10;
unsigned int num_antigen_producers_1 = 10;
unsigned int antigen_producer_action_time = 100;
unsigned int num_cells_2 = 50;
unsigned int cell_lifespan_2 = 100;
unsigned int num_cell_receptors_2 = 10;
unsigned int num_vr_receptors_2 = 10;
unsigned int num_response_producers_2 = 10;
unsigned int ts_port = TS_DEFAULT_PORT;
unsigned int log_tissue = 0;
unsigned int probe_rate = 100000;
unsigned int antigen_multiplier = 10;

char parameter_file[64] = "";

/* monitor thymus periodically + save data to log */
void probe_thymus(Tissue *tissue)
{
	register unsigned int i;
	static char message[1024], sc_name[64];

	for(i = 0; i < tissue->num_cells; i++) {
		/* save cytokine receptor and antigen producer action time */
		if(tissue->cell[i] && tissue->cell[i]->type == 1) {
			snprintf(message, sizeof(message) - 1, "cr0 %d %f %d", \
					tissue->cell[i]->id, \
					tissue->cell[i]->cytokine_receptor[0]->val, \
					tissue->cell[i]->antigen_producer[0]->action_time);
			logtime_tissue(message);
			continue;
		}
		/* save vr receptor repertoire + cytokine rate */
		if(tissue->cell[i] && tissue->cell[i]->type == 2) {
			snprintf(message, sizeof(message) - 1, "vrr %d %s %f", \
					tissue->cell[i]->id, \
					syscall_name(antigentoi(tissue->cell[i]->vr_receptor[0]->lock), sc_name), \
					tissue->cell[i]->cytokine[0]->val);
			logtime_tissue(message);
			continue;
		}
	}
}

/* create antigen receptor */
Antigen *random_lock(void)
{
	Antigen *lock = NULL;
	static char line[64];
	static unsigned val;

	/* set lock to random syscall number */
	val = (int) (((double) nsyscalls * rand()) / (RAND_MAX + 1.0));
	snprintf(line, sizeof(line), "%d", val);
	lock = set_antigen(lock, line);

	return lock;
}

/* produce a response */
char *response_message(Cell *cell, unsigned int ind)
{
	Response_Producer *rp;
	static char line[256], name[256];
	char *message;

	rp = cell->response_producer[ind];
	/* create policy */
	if(rp->type == 1) {  /* type 1 response = add permit policy */
		snprintf(line, sizeof(line), "%d %d %s(%d) permit", \
				cell->id, 
				rp->antigen->id, \
				syscall_name(antigentoi(rp->antigen), name), \
				antigentoi(rp->antigen));
	} else
		err(1, "unknown response producer type");

	/* set message */
	if(!(message = calloc(strlen(line) + 1, sizeof(char))))
		err(1, "couldn't allocate memory for message");
	strncpy(message, line, strlen(line) + 1);

	return message;
}


/* cell cycle for thymus cell type 1 (dc) */
void cell_cycle_1(Cell *cell)
{
	unsigned int i;

	/* set antigen producer action time to be responsive to a signal */
	if(cell->num_cytokine_receptors == 1)
		if(cell->cytokine_receptor[0]->val != 0.0) {
			if(cell->cytokine_receptor[0]->val > 0.0)
				for(i = 0; i < cell->num_antigen_producers; i++)
					cell->antigen_producer[i]->action_time = \
							antigen_producer_action_time;
			else
				for(i = 0; i < cell->num_antigen_producers; i++)
					cell->antigen_producer[i]->action_time *= 0.5;
		}

	/* and just collect + present antigen... */
}

/* cell cycle for thymus cell type 2 (tcell) */
void cell_cycle_2(Cell *cell)
{
	unsigned int i, j, ind;
	Antigen *lock;

	/* replace vr receptor periodically */
	if(cell_lifespan_2 && cell->iterations >= cell_lifespan_2 && \
			cell->cytokine[0]->val == 0.0) {  /* not 0.0 = previous match */
		lock = random_lock();
		for(i = 0; i < num_vr_receptors_2; i++) {
			free_antigen(cell->vr_receptor[i]->lock);
			cell->vr_receptor[i]->lock = dup_antigen(lock);
		}
		cell->iterations = 0;
	}

	/* are we in contact with a cell of type 1 */
	for(i = 0; i < cell->num_cell_receptors; i++)
		if(cell->cell_receptor[i]->cell)
			if((cell->cell_receptor[i]->cell->type == 1))
				/* any antigen on vr receptor */
				for(j = 0; j < num_vr_receptors_2; j++)
					if(cell->vr_receptor[j]->key) {
						/* yes, so produce response to antigen */
						/* choose random response producer */
						ind = (int) (((double) num_response_producers_2 \
								* rand()) / (RAND_MAX + 1.0));
						/* respond to matched antigen */
						set_rp_antigen(cell->response_producer[ind], \
								cell->vr_receptor[j]->key);
						if(cell->response_producer[ind]->response)
							free(cell->response_producer[ind]->response);
						cell->response_producer[ind]->response = \
								response_message(cell, ind);
						cell->cytokine[0]->val += 1.0;  /* count matches */
				}
}

/* initialise some cells */
void create_thymus_cells(Tissue *thymus)
{
	Cell_Param cell_param;
	unsigned int i, j, ind;
	Antigen *lock;

	i = num_cells_1;
	while(i) {
		do  /* find spare slot in thymus to put cell */
			ind = rand() % thymus->num_cells;
		while(thymus->cell[ind]);

		/* set parameters for cell creation */
		bzero(&cell_param, sizeof(Cell_Param));
		cell_param.id = ind;
		cell_param.type = 1;  /* type 1 */
		cell_param.num_antigen = num_antigen_1; /* num of antigen cell stored */
		cell_param.num_cytokine_receptors = 1;  /* respond to signal */
		cell_param.num_antigen_receptors = num_antigen_receptors_1;
		cell_param.num_antigen_producers = num_antigen_producers_1;
		cell_param.cycle = &cell_cycle_1;

		/* initialise the cell */
		thymus->cell[ind] = init_cell(&cell_param);

		/* adjust any parameters */
		/* set cytokine receptor to respond to signal 2 (ID 1) */
		thymus->cell[ind]->cytokine_receptor[0]->id = 1;
		/* set action times for antign producers */
		for(j = 0; j < num_antigen_producers_1; j++)
			thymus->cell[ind]->antigen_producer[j]->action_time = \
					antigen_producer_action_time;

		i--;  /* cell created */
	}

	i = num_cells_2;
	while(i) {
		do  /* find spare slot in thymus to put cell */
			ind = rand() % thymus->num_cells;
		while(thymus->cell[ind]);

		/* set parameters for cell creation */
		bzero(&cell_param, sizeof(Cell_Param));
		cell_param.id = ind;
		cell_param.type = 2;  /* type 2 */
		cell_param.num_cytokines = 1;  /* keep a count of num of matches */
		cell_param.num_cell_receptors = num_cell_receptors_2;
		cell_param.num_vr_receptors = num_vr_receptors_2;
		cell_param.num_response_producers = num_response_producers_2;
		cell_param.cycle = &cell_cycle_2;

		/* initialise the cell */
		thymus->cell[ind] = init_cell(&cell_param);

		/* adjust any parameters */
		for(j = 0; j < num_cell_receptors_2; j++)
			thymus->cell[ind]->cell_receptor[j]->type = 1;
		for(j = 0; j < num_response_producers_2; j++)
			thymus->cell[ind]->response_producer[j]->type = 1;
		/* set random values for locks of vr receptors */
		lock = random_lock();  /* same lock for all receptors */
		for(j = 0; j < num_vr_receptors_2; j++) {
			thymus->cell[ind]->vr_receptor[j]->type = 1;
			thymus->cell[ind]->vr_receptor[j]->state = 1;
			thymus->cell[ind]->vr_receptor[j]->lock = dup_antigen(lock);
			thymus->cell[ind]->vr_receptor[j]->threshold = 1.0;
			thymus->cell[ind]->vr_receptor[j]->match = &antigen_match_syscall;
		}
		/* set cell iterations to random value (to avoid lockstep */
		thymus->cell[ind]->iterations = (int) (((double) \
				cell_lifespan_2 * rand()) / (RAND_MAX + 1.0));
	
		i--;  /* cell created */
	}
}

/* read parameters from file */
void read_parameters(char *filename)
{
	read_parameter_uint(filename, "max_antigen", &max_antigen);
	read_parameter_uint(filename, "max_cytokines", &max_cytokines);
	read_parameter_uint(filename, "max_cells", &max_cells);
	read_parameter_uint(filename, "cell_update_rate", &cell_update_rate);
	read_parameter_uint(filename, "num_cells_1", &num_cells_1);
	read_parameter_uint(filename, "num_antigen_1", &num_antigen_1);
	read_parameter_uint(filename, "num_antigen_receptors_1", \
			&num_antigen_receptors_1);
	read_parameter_uint(filename, "num_antigen_producers_1", \
			&num_antigen_producers_1);
	read_parameter_uint(filename, "antigen_producer_action_time", \
			&antigen_producer_action_time);
	read_parameter_uint(filename, "num_cells_2", &num_cells_2);
	read_parameter_uint(filename, "cell_lifespan_2", &cell_lifespan_2);
	read_parameter_uint(filename, "num_cell_receptors_2", \
			&num_cell_receptors_2);
	read_parameter_uint(filename, "num_vr_receptors_2", &num_vr_receptors_2);
	read_parameter_uint(filename, "num_response_producers_2", \
			&num_response_producers_2);
	read_parameter_uint(filename, "ts_port", &ts_port);
	read_parameter_uint(filename, "log_tissue", &log_tissue);
	read_parameter_uint(filename, "probe_rate", &probe_rate);
	read_parameter_uint(filename, "antigen_multiplier", &antigen_multiplier);
}

static void parse_command_line(int argc, char **argv)
{
	if(argc == 2) {
		strncpy(parameter_file, argv[1], sizeof(parameter_file));
		read_parameters(parameter_file);
	}
}

int main(int argc, char **argv)
{
	Tissue *thymus;

	parse_command_line(argc, argv);

	/* create some tissue - a thymus */
	thymus = init_tissue((int16_t) ts_port, max_antigen, max_cytokines, \
			max_cells);

	/* save user specified parameters to log */
	log_parameter_file(parameter_file);

	/* log input data (from tissue clients) to thymus */
	thymus->log = (short int) log_tissue;

	/* set multiplier for incoming antigen */
	thymus->antigen_multiplier = antigen_multiplier;

	/* initialise some cells in the thymus */
	create_thymus_cells(thymus);

	/* initialise a probe to monitor thymus */
	init_probe(thymus, probe_thymus, probe_rate);

	/* run positive selection algorithm */
	while(1) {
		/* let cells interact in tissue */
		step_tissue(thymus);

		/* wait a bit */
		usleep(cell_update_rate);
	}

	return EXIT_SUCCESS;
}
